/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_api_unix.cpp
 *
 * Purpose:     UNIX implementation file for the recls API.
 *
 * Created:     8th September 2003
 * Updated:     25th September 2003
 *
 * Copyright:   Synesis Software Pty Ltd, (c) 2003. All rights reserved.
 *
 * Home:        http://www.synesis.com.au/software
 *              http://recls.org/
 *
 * ////////////////////////////////////////////////////////////////////////// */

#error This file is not in the least finished, so don't even try to compile it

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include "recls.h"
#include "recls_internal.h"
#include "recls_assert.h"
#include "recls_util.h"

#include <stlsoft_nulldef.h>

#include <unixstl_readdir_sequence.h>
#include <unixstl_findfile_sequence.h>

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

typedef unixstl_ns_qual(readdir_sequence)       readdir_sequence_t;
typedef unixstl_ns_qual(findfile_sequence)      findfile_sequence_t;

/* ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * UnixDirectoryNode
 */

class UnixDirectoryNode
    : public ReclsDNode
{
// Construction
private:
    UnixDirectoryNode(recls_uint32_t flags, char const *rootDir, char const *pattern);
public:
    virtual ~UnixDirectoryNode();

    static UnixDirectoryNode *FindAndCreate(recls_uint32_t flags, char const *rootDir, char const *pattern);

// ReclsDNode methods
private:

    virtual recls_rc_t GetNext();

    virtual recls_rc_t GetDetails(recls_info_t *pinfo);

    virtual recls_rc_t GetNextDetails(recls_info_t *pinfo);

// Implementation
private:
    recls_rc_t      Initialise();

    recls_bool_t    _is_valid() const;

// Members
private:
    recls_info_t                        m_current;
    ReclsDNode                          *m_dnode;
    recls_uint32_t const                m_flags;
    findfile_sequence_t                 m_files;
    findfile_sequence_t::const_iterator m_filesBegin;
    readdir_sequence_t                  m_directories;
    readdir_sequence_t::const_iterator  m_directoriesBegin;
    char                                m_rootDir[RECLS_PATH_MAX];
    char                                m_pattern[RECLS_PATH_MAX];
};

UnixDirectoryNode::UnixDirectoryNode(recls_uint32_t flags, char const *rootDir, char const *pattern)
    : m_flags(flags)
    , m_files(rootDir, pattern)
    , m_filesBegin(m_files.begin())
    , m_directories(rootDir, "*.*", findfile_sequence_t::onlyDir)
    , m_directoriesBegin((flags & RECLS_F_RECURSIVE) ? m_directories.begin() : m_directories.end())
    , m_current(NULL)
    , m_dnode(NULL)
{
    strcpy(m_rootDir, rootDir);
    strcpy(m_pattern, pattern);

    recls_assert(stlsoft_raw_offsetof(UnixDirectoryNode, m_files) < stlsoft_raw_offsetof(UnixDirectoryNode, m_filesBegin));
    recls_assert(stlsoft_raw_offsetof(UnixDirectoryNode, m_directories) < stlsoft_raw_offsetof(UnixDirectoryNode, m_directoriesBegin));
}

inline /* static */ UnixDirectoryNode *UnixDirectoryNode::FindAndCreate(recls_uint32_t flags, char const *rootDir, char const *pattern)
{
    UnixDirectoryNode   *node;

#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
    try
    {
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */
        node = new UnixDirectoryNode(flags, rootDir, pattern);
#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
    }
    catch(std::bad_alloc &)
    {
        node = NULL;
    }
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */

    if(NULL != node)
    {
        // Ensure that it, or one of its sub-nodes, has matching entries.
        recls_rc_t  rc = node->Initialise();

        if(RECLS_FAILED(rc))
        {
            delete node;

            node = NULL;
        }
    }

    recls_assert(NULL == node || node->_is_valid());

    return node;
}

UnixDirectoryNode::~UnixDirectoryNode()
{
    FileInfo_Release(m_current);

    delete m_dnode;
}

recls_rc_t UnixDirectoryNode::Initialise()
{
    recls_rc_t  rc;

    recls_assert(NULL == m_current);
    recls_assert(NULL == m_dnode);

    if(m_filesBegin != m_files.end())
    {
        // (i) Try getting a file first,
        m_current = create_info_from_W32FD(m_rootDir, (*m_filesBegin).get_find_data());

        if(NULL == m_current)
        {
            rc = RECLS_RC_OUT_OF_MEMORY;
        }
        else
        {
            rc = RECLS_RC_OK;
        }
    }
    else
    {
        if(m_directoriesBegin == m_directories.end())
        {
            rc = RECLS_RC_NO_MORE_DATA;
        }
        else
        {
            do
            {
                m_dnode = UnixDirectoryNode::FindAndCreate(m_flags, (*m_directoriesBegin).get_path(), m_pattern);

            } while(NULL == m_dnode && ++m_directoriesBegin != m_directories.end());

            rc = (NULL == m_dnode) ? RECLS_RC_NO_MORE_DATA : RECLS_RC_OK;
        }
    }

    if(RECLS_SUCCEEDED(rc))
    {
        recls_assert(_is_valid());
    }

    return rc;
}

recls_bool_t UnixDirectoryNode::_is_valid() const
{
    recls_rc_t  rc  =   RECLS_RC_OK;

    if(RECLS_SUCCEEDED(rc))
    {

    }

    if(RECLS_SUCCEEDED(rc))
    {
    }

    if(RECLS_SUCCEEDED(rc))
    {
    }

    if(RECLS_SUCCEEDED(rc))
    {
    }

    // (i) Either we are enumerating files (m_current != NULL) or directories (m_dnode != NULL), but not both
    recls_assert(NULL == m_current || NULL == m_dnode);
    // (ii) Either we are enumerating files (m_current != NULL) or there are no more files to enumerate
    recls_assert(NULL != m_current || m_filesBegin == m_files.end());
    /// (iii) Either we are enumerating
//  recls_assert(NULL != m_dnode || m_directoriesBegin == m_directories.end());

    return RECLS_SUCCEEDED(rc);
}


recls_rc_t UnixDirectoryNode::GetNext()
{
    recls_assert(_is_valid());

    /* Searching operates as follows:
     *
     * 1. Return all the contents of the files
     * 2. Return the contents of the directories.
     *
     * Hence, if m_dnode is non-NULL, we've already searched
     */

    /* States:
     *
     * - Iterating files: m_filesBegin != m_files.end(), m_dnode is NULL, m_directoriesBegin != m_directories.end()
     * - Iterating directories: m_directoriesBegin != m_directories.end(), m_dnode is non-NULL, m_current is NULL
     *
     */

    // Invariants

    // (i) Either we are enumerating files (m_current != NULL) or directories (m_dnode != NULL), but not both
    recls_assert(NULL == m_current || NULL == m_dnode);
    // (ii) Either we are enumerating files (m_current != NULL) or there are no more files to enumerate
    recls_assert(NULL != m_current || m_filesBegin == m_files.end());
    /// (iii) Either we are enumerating
//  recls_assert(NULL != m_dnode || m_directoriesBegin == m_directories.end());

    recls_rc_t  rc = RECLS_RC_NO_MORE_DATA;

    if(NULL != m_current)
    {
        // Currently enumerating through the files

        recls_assert(m_filesBegin != m_files.end());
        recls_assert(NULL == m_dnode);

        // Advance, and check for end of sequence
        ++m_filesBegin;

        FileInfo_Release(m_current);
        if(m_filesBegin != m_files.end())
        {
            // Still enumerating, so just update m_current
            m_current = create_info_from_W32FD(m_rootDir, (*m_filesBegin).get_find_data());

            rc = RECLS_RC_OK;
        }
        else
        {
            // No more left in the files sequence, so delete m_current
            m_current = NULL;

            rc = RECLS_RC_NO_MORE_DATA;
        }
    }

    if(NULL == m_current)
    {
        // Now we are either enumerating the directories, or we've already done so

        if(NULL != m_dnode)
        {
            // Currently enumerating the directories
            rc = m_dnode->GetNext();

            if(RECLS_RC_NO_MORE_DATA == rc)
            {
                ++m_directoriesBegin;

                delete m_dnode;

                m_dnode = NULL;
            }
        }

        if(m_directoriesBegin == m_directories.end())
        {
            // Enumeration is complete.
            rc = RECLS_RC_NO_MORE_DATA;
        }
        else
        {
            if(NULL == m_dnode)
            {
                do
                {
                    // Creation of the node will cause it to enter the first enumeration
                    // state. However, if there are no matching

                    recls_assert(m_directoriesBegin != m_directories.end());

                    m_dnode = UnixDirectoryNode::FindAndCreate(m_flags, (*m_directoriesBegin).get_path(), m_pattern);

                    if(NULL != m_dnode)
                    {
                        rc = RECLS_RC_OK;
                    }
                    else
                    {
                        ++m_directoriesBegin;
                    }

                } while(NULL == m_dnode && m_directoriesBegin != m_directories.end());
            }
        }
    }

    recls_assert(_is_valid());

    return rc;
}

recls_rc_t UnixDirectoryNode::GetDetails(recls_info_t *pinfo)
{
    recls_assert(_is_valid());

    recls_rc_t  rc;

    recls_assert(NULL != pinfo);
    recls_assert(NULL == m_current || NULL == m_dnode);

    if(NULL != m_current)
    {
        // Currently searching for files from the current directory

        recls_assert(NULL == m_dnode);

        rc = FileInfo_Copy(m_current, pinfo);
    }
    else if(NULL != m_dnode)
    {
        recls_assert(NULL == m_current);

        // Sub-directory searching is active, so get from there.

        rc = m_dnode->GetDetails(pinfo);
    }
    else
    {
        // Enumeration has completed
        rc = RECLS_RC_NO_MORE_DATA;
    }

    recls_assert(_is_valid());

    return rc;
}

recls_rc_t UnixDirectoryNode::GetNextDetails(recls_info_t *pinfo)
{
    recls_assert(_is_valid());
    recls_assert(NULL != pinfo);

    recls_rc_t  rc  =   GetNext();

    if(RECLS_SUCCEEDED(rc))
    {
        rc = GetDetails(pinfo);
    }

    recls_assert(_is_valid());

    return rc;
}

/* /////////////////////////////////////////////////////////////////////////////
 * ReclsSearchInfo
 */

void *ReclsSearchInfo::operator new(size_t cb, int cDirParts, size_t cbRootDir, size_t cbPattern)
{
    cbRootDir = align_up_size(cbRootDir);
    cbPattern = align_up_size(cbPattern);

    recls_assert(cb > stlsoft_raw_offsetof(ReclsSearchInfo, data));

    cb = stlsoft_raw_offsetof(ReclsSearchInfo, data)
       + (cDirParts) * sizeof(recls_strptrs_t)
       + cbRootDir
       + cbPattern;

    return malloc(cb);
}

#if !defined(RECLS_COMPILER_IS_BORLAND) && \
    !defined(RECLS_COMPILER_IS_DMC)
void ReclsSearchInfo::operator delete(void *pv, int cDirParts, size_t cbRootDir, size_t cbPattern)
{
    free(pv);
}
#endif /* !RECLS_COMPILER_IS_BORLAND && !RECLS_COMPILER_IS_DMC */

void ReclsSearchInfo::operator delete(void *pv)
{
    free(pv);
}

inline /* static */ recls_rc_t ReclsSearchInfo::FindAndCreate(char const *rootDir, char const *pattern, recls_uint32_t flags, ReclsSearchInfo **ppsi)
{
    recls_rc_t      rc;
    ReclsSearchInfo *si;

    char    fullPath[RECLS_PATH_MAX];
    char    *_p;

    *ppsi = NULL;

    if(0 == ::GetFullPathNameA(rootDir, RECLS_NUM_ELEMENTS(fullPath), fullPath, &_p))
    {
        rc = RECLS_RC_INVALID_DIRECTORY;
    }
    else
    {
        size_t  lenSearchRoot   =   strlen(fullPath);
        size_t  lenPattern      =   strlen(pattern);

        recls_assert(0 < lenSearchRoot);

        if(fullPath[lenSearchRoot - 1] != '\\')
        {
            fullPath[lenSearchRoot] = '\\';
            fullPath[lenSearchRoot + 1] = '\0';

            ++lenSearchRoot;
        }

        rootDir = fullPath;

        // Count the directory parts.
        char const          *begin      =   rootDir + 2;
        char const *const   end         =   rootDir + lenSearchRoot;
        int                 cDirParts   =   count_dir_parts(begin, end);

#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
        try
        {
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */
            si = new(cDirParts, sizeof(char) * (1 + lenSearchRoot), sizeof(char) * (1 + lenPattern)) ReclsSearchInfo(cDirParts, rootDir, pattern, flags);
#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
        }
        catch(std::bad_alloc &)
        {
            si = NULL;
        }
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */

        if(NULL == si)
        {
            rc = RECLS_RC_FAIL;
        }
        else
        {
            // This is a nasty hack. It's tantamount to ctor & create function, so
            // should be made more elegant soon.
            if(NULL == si->m_dnode)
            {
                delete si;

                si = NULL;

                rc = RECLS_RC_NO_MORE_DATA;
            }
            else
            {
                *ppsi = si;

                rc = RECLS_RC_OK;
            }
        }
    }

    return rc;
}


char const *ReclsSearchInfo::_calc_rootDir(int cDirParts, char const *rootDir)
{
    // Root dir is located after file parts, and before pattern
    return strcpy((char*)&data[cDirParts * sizeof(recls_strptrs_t)], rootDir);
}

char const *ReclsSearchInfo::_calc_pattern(int cDirParts, char const *rootDir, char const *pattern)
{
    // Pattern is located after file parts and root dir
    size_t  cb = align_up_size((1 + strlen(rootDir)) * sizeof(char));

    return strcpy((char*)&data[cDirParts * sizeof(recls_strptrs_t) + cb], pattern);
}

ReclsSearchInfo::ReclsSearchInfo(   int             cDirParts
                                ,   char const      *rootDir
                                ,   char const      *pattern
                                ,   recls_uint32_t  flags)
    : m_flags(flags)
    , m_lastError(RECLS_RC_OK)
    , m_rootDir(_calc_rootDir(cDirParts, rootDir))
    , m_pattern(_calc_pattern(cDirParts, rootDir, pattern))
{
    recls_assert(NULL != rootDir);
    recls_assert(NULL != pattern);
    recls_assert(strlen(rootDir) < RECLS_PATH_MAX);
    recls_assert(strlen(pattern) < RECLS_PATH_MAX);

    // Initialise the directory parts.

    recls_assert(rootDir[1] == ':');

    char const              *p          =   rootDir + 2;
    struct recls_strptrs_t  *begin      =   (struct recls_strptrs_t*)&data[0];

    m_directoryParts.begin              =   begin;
    m_directoryParts.end                =   begin + cDirParts;

    begin->begin = p;

    for(; *p; ++p)
    {
        if(*p == '\\')
        {
            begin->end = p + 1;

            if(++begin != m_directoryParts.end)
            {
                begin->begin = p + 1;
            }
        }
    }

    // Now start the search
    m_dnode = UnixDirectoryNode::FindAndCreate(m_flags, rootDir, pattern);
}

// Operations
recls_rc_t ReclsSearchInfo::GetNext()
{
    recls_assert(NULL != m_dnode);

    m_lastError =   m_dnode->GetNext();

    if(RECLS_RC_NO_MORE_DATA == m_lastError)
    {
        delete m_dnode;

        m_dnode = NULL;
    }

    return m_lastError;
}

recls_rc_t ReclsSearchInfo::GetDetails(recls_info_t *pinfo)
{
    recls_assert(NULL != m_dnode);

    return (m_lastError = m_dnode->GetDetails(pinfo));
}

recls_rc_t ReclsSearchInfo::GetNextDetails(recls_info_t *pinfo)
{
    recls_assert(NULL != m_dnode);

    m_lastError =   m_dnode->GetNextDetails(pinfo);

    if(RECLS_RC_NO_MORE_DATA == m_lastError)
    {
        delete m_dnode;
    }

    return m_lastError;
}

// Accessors

recls_rc_t ReclsSearchInfo::GetLastError() const
{
    return m_lastError;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Search control
 */

/**
 *
 *
 * \param
 *
 */
RECLS_FNDECL(recls_rc_t) Recls_Search(  char const      *searchRoot
                                    ,   char const      *pattern
                                    ,   recls_uint32_t  flags
                                    ,   hrecls_t        *phSrch)
{
    recls_assert(NULL != searchRoot);
    recls_assert(NULL != pattern);
    recls_assert(NULL != phSrch);

    *phSrch = ReclsSearchInfo::ToHandle(NULL);

    recls_rc_t  rc;

    // Validate the search root
    if( NULL == searchRoot ||
        0 == *searchRoot)
    {
        searchRoot = ".";
    }

    // Validate the pattern.
    if('\0' == *pattern)
    {
        rc = RECLS_RC_NO_MORE_DATA;
    }
    else
    {
        ReclsSearchInfo *si;

        rc = ReclsSearchInfo::FindAndCreate(searchRoot, pattern, flags, &si);

        if(RECLS_SUCCEEDED(rc))
        {
            *phSrch = ReclsSearchInfo::ToHandle(si);

            rc = RECLS_RC_OK;
        }
    }

    return rc;
}

RECLS_FNDECL(recls_rc_t) Recls_SearchProcess(   char const                      *searchRoot
                                            ,   char const                      *pattern
                                            ,   recls_uint32_t                  flags
                                            ,   hrecls_process_fn_t             pfn
                                            ,   recls_process_fn_param_t        param)
{
    return RECLS_RC_NOT_IMPLEMENTED;
}

/** Closes the given search */
RECLS_FNDECL(void) Recls_SearchClose(hrecls_t hSrch)
{
    ReclsSearchInfo *si =   ReclsSearchInfo::FromHandle(hSrch);

    recls_assert(NULL != si);

    delete si;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Search enumeration
 */

RECLS_FNDECL(recls_rc_t) Recls_GetNext(hrecls_t hSrch)
{
    ReclsSearchInfo *si =   ReclsSearchInfo::FromHandle(hSrch);

    recls_assert(NULL != si);

    return si->GetNext();
}

RECLS_FNDECL(recls_rc_t) Recls_GetNextDetails(  hrecls_t        hSrch
                                            ,   recls_info_t    *pinfo)
{
    ReclsSearchInfo *si =   ReclsSearchInfo::FromHandle(hSrch);

    recls_assert(NULL != si);

    return si->GetNextDetails(pinfo);
}

RECLS_FNDECL(void) Recls_CloseDetails(recls_info_t fileInfo)
{
    recls_assert(NULL != fileInfo);

    FileInfo_Release(fileInfo);
}

RECLS_FNDECL(recls_rc_t) Recls_CopyDetails( recls_info_t    fileInfo
                                        ,   recls_info_t    *pinfo)
{
    recls_assert(NULL != fileInfo);

    return FileInfo_Copy(fileInfo, pinfo);
}

RECLS_FNDECL(recls_rc_t) Recls_OutstandingDetails(  hrecls_t        hSrch
                                                    ,   recls_uint32_t  *count)
{
    ReclsSearchInfo *si =   ReclsSearchInfo::FromHandle(hSrch);

    recls_assert(NULL != si);
    recls_assert(NULL != count);

    recls_sint32_t  cCreated;
    recls_sint32_t  cShared;

    FileInfo_BlockCount(&cCreated, &cShared);

    *count = cCreated;

    return RECLS_RC_OK;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Error handling
 */

RECLS_FNDECL(int) Recls_GetErrorString( recls_rc_t  rc
                                    ,   char        *buffer
                                    ,   size_t      cchBuffer)
{
    recls_assert(NULL != buffer);

    if(rc == RECLS_RC_SEARCH_NO_CURRENT)
    {
        strncpy(buffer, "Search has no current node", cchBuffer);
    }
    else if(rc == RECLS_RC_INVALID_DIRECTORY)
    {
        strncpy(buffer, "Invalid directory", cchBuffer);
    }
    else if(rc == RECLS_RC_NO_MORE_DATA)
    {
        strncpy(buffer, "No more data", cchBuffer);
    }
    else if(rc == RECLS_RC_OUT_OF_MEMORY)
    {
        strncpy(buffer, "No more memory", cchBuffer);
    }

    return strlen(buffer);
}

/* /////////////////////////////////////////////////////////////////////////////
 * Property elicitation
 */

RECLS_FNDECL(size_t) Recls_GetDirectoryPathProperty(    recls_info_t    fileInfo
                                                    ,   char            *buffer
                                                    ,   size_t          cchBuffer)
{
    return Recls_GetDirectoryProperty(fileInfo, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetShortFileProperty(    recls_info_t    fileInfo
                                                ,   char            *buffer
                                                ,   size_t          cchBuffer)
{
    return Recls_GetFileProperty(fileInfo, buffer, cchBuffer);
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileReadOnly(recls_info_t fileInfo)
{
    recls_assert(NULL != fileInfo);

    return fileInfo->attributes & ????;
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileDirectory(recls_info_t fileInfo)
{
    recls_assert(NULL != fileInfo);

    return fileInfo->attributes & ????;
}

RECLS_FNDECL(recls_time_t) Recls_GetCreationTime(recls_info_t fileInfo)
{
    recls_assert(NULL != fileInfo);

    return fileInfo->modificationTime;
}

RECLS_FNDECL(recls_time_t) Recls_GetLastStatusChangeTime(recls_info_t fileInfo)
{
    recls_assert(NULL != fileInfo);

    return fileInfo->lastStatusChangeTime;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
