/* /////////////////////////////////////////////////////////////////////////////
 * File:        reclstl_search_sequence.h
 *
 * Purpose:     Contains the basic_search_sequence template class, and ANSI
 *              and Unicode specialisations thereof.
 *
 * Created:     10th September 2003
 * Updated:     13th October 2003
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/software
 *                          http://www.recls.org/
 *
 *              email:      submissions@recls.org  for submissions
 *                          admin@recls.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef RECLS_INCL_H_WINSTL_SEARCH_SEQUENCE
#define RECLS_INCL_H_WINSTL_SEARCH_SEQUENCE

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_H_WINSTL_SEARCH_SEQUENCE_MAJOR       1
# define RECLS_VER_H_WINSTL_SEARCH_SEQUENCE_MINOR       1
# define RECLS_VER_H_WINSTL_SEARCH_SEQUENCE_REVISION    8
# define RECLS_VER_H_WINSTL_SEARCH_SEQUENCE_EDIT        11
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/** \file reclstl_search_sequence.h Definition of the \link recls::stl::basic_search_sequence basic_search_sequence\endlink 
 * class, and supporting classes 
 * (\link recls::stl::basic_search_sequence_const_iterator basic_search_sequence_const_iterator\endlink and 
 * \link recls::stl::basic_search_sequence_value_type basic_search_sequence_value_type\endlink), representing the
 * \ref group_recls_stl "recls STL mapping" */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef RECLS_INCL_H_RECLSTL
# include "reclstl.h"
#endif /* !RECLS_INCL_H_RECLSTL */
#ifndef RECLS_INCL_H_RECLS_ASSERT
# include <recls_assert.h>
#endif /* !RECLS_INCL_H_RECLS_ASSERT */
#ifndef _STLSOFT_INCL_H_STLSOFT_ITERATOR
# include <stlsoft_iterator.h>
#endif /* !_STLSOFT_INCL_H_STLSOFT_ITERATOR */

#ifndef _STLSOFT_INCL_H_STLSOFT_PROXY_SEQUENCE
# include <inprogress/stlsoft_proxy_sequence.h>
#endif /* !_STLSOFT_INCL_H_STLSOFT_PROXY_SEQUENCE */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{

namespace stl
{
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Forward declarations
 */

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION

template<   typename C
        ,   typename T
        >
class basic_search_sequence_value_type;

template<   typename C
        ,   typename T
        ,   typename V
        >
class basic_search_sequence_const_iterator;

#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// \brief A traits class to provide character-encoding specific mappings from the \ref group_recls "recls C API" to the \ref group_recls_stl "recls.stl" mapping"
/// \ingroup group_recls_stl
/// 
template <typename C>
struct reclstl_traits
{
public:
    /// The <code>char</code> type
    ///
    /// \note In this template, <code>void</code> is a placeholder. <code>char</code> and <code>wchar_t</code> specialisations
    /// will use their respective character types
    typedef void            char_type;
    /// The entry type
    ///
    /// \note In this template, <code>void*</code> is a placeholder. <code>char</code> and <code>wchar_t</code> specialisations
    /// will use their respective \link recls::recls_fileinfo_t recls_fileinfo_t\endlink types
    typedef void            *entry_type;

public:
    /// traits-form of \link recls::Recls_Search Recls_Search\endlink
    static hrecls_t         Search(char_type const *searchRoot, char_type const *pattern, recls_uint32_t flags);

    /// traits-form of \link recls::Recls_GetDetails Recls_GetDetails\endlink
    static recls_rc_t       GetDetails(hrecls_t hSrch, entry_type *pinfo);
    /// traits-form of \link recls::Recls_GetNextDetails Recls_GetNextDetails\endlink
    static recls_rc_t       GetNextDetails(hrecls_t hSrch, entry_type *pinfo);

    /// traits-form of \link recls::Recls_CloseDetails Recls_CloseDetails\endlink
    static void             CloseDetails(entry_type fileInfo);
    /// traits-form of \link recls::Recls_CopyDetails Recls_CopyDetails\endlink
    static entry_type       CopyDetails(entry_type fileInfo);

    /// traits-form of <code>strcpy</code>
    static char_type        *str_copy(char_type *, char_type const *);
};

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
template <>
struct reclstl_traits<char>
{
public:
    typedef reclstl_traits<char>                traits_type;
    typedef char                                char_type;
    typedef recls_info_t                        entry_type;
    typedef recls_strptrs_t                     strptrs_type;
    typedef recls_strptrsptrs_t                 strptrsptrs_type;

    /// The directory parts sequence type
    typedef stlsoft::proxy_sequence<    const strptrs_type
                                    ,   string_t
                                    ,   traits_type // Note, this type also serves as the traits type for the STLSoft proxy_sequence
                                    >           directory_parts_type;

    /// Constructs a value of type \c string_t from an instance of the pointers structure.
    ///
    /// \note This is required by the STLSoft proxy_sequence template.
    static string_t make_value(strptrs_type const &ptrs)
    {
        return string_t(ptrs.begin, ptrs.end);
    }

public:
    static hrecls_t Search(char_type const *searchRoot, char_type const *pattern, recls_uint32_t flags)
    {
        hrecls_t    hSrch;
        recls_rc_t  rc = Recls_Search(searchRoot, pattern, flags, &hSrch);

        return RECLS_SUCCEEDED(rc) ? hSrch : static_cast<hrecls_t>(NULL);
    }

    static recls_rc_t GetDetails(hrecls_t hSrch, entry_type *pinfo)
    {
        return Recls_GetDetails(hSrch, pinfo);
    }

    static recls_rc_t GetNextDetails(hrecls_t hSrch, entry_type *pinfo)
    {
        return Recls_GetNextDetails(hSrch, pinfo);
    }

    static void CloseDetails(entry_type fileInfo)
    {
        Recls_CloseDetails(fileInfo);
    }
    static entry_type CopyDetails(entry_type fileInfo)
    {
        entry_type  infoCopy;

        return RECLS_SUCCEEDED(Recls_CopyDetails(fileInfo, &infoCopy)) ? infoCopy : static_cast<entry_type>(NULL);
    }

    static char_type *str_copy(char_type *dest, char_type const *src)
    {
        return strcpy(dest, src);
    }
};
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

// class basic_search_sequence
/// Presents an STL-like sequence interface over the items on the file-system
///
/// \ingroup group_recls_stl
///
/// \param C The character type
/// \param T The traits type. On translators that support default template arguments this defaults to reclstl_traits<C>
template<   typename C
#ifdef __STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_CLASS_ARGUMENT_SUPPORT
        ,   typename T = reclstl_traits<C>
#else
        ,   typename T /* = reclstl_traits<C> */
#endif /* __STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_CLASS_ARGUMENT_SUPPORT */
        >
class basic_search_sequence
{
/// \name Types
/// @{
public:
    /// The character type
    typedef C                                                       char_type;
    /// The traits type
    typedef T                                                       traits_type;
    /// The current parameterisation of the type
    typedef basic_search_sequence<C, T>                             class_type;
    /// The value type
    typedef basic_search_sequence_value_type<C, T>                  value_type;
    /// The non-mutating (const) iterator type supporting the Input Iterator concept
    typedef basic_search_sequence_const_iterator<C, T, value_type>  const_iterator;
    /// The reference type
    typedef value_type                                              &reference;
    /// The non-mutable (const) reference type
    typedef value_type const                                        &const_reference;
    /// The find-data type
    typedef ss_typename_type_k traits_type::entry_type              entry_type;
    /// The size type
    typedef size_t                                                  size_type;

/// @}

/// \name Construction
/// @{
public:
    /// Commence a search according to the given search pattern and flags 
    basic_search_sequence(char_type const *searchSpec, recls_uint32_t flags);
    /// Commence a search according to the given search pattern and flags, relative to \c directory
    basic_search_sequence(char_type const *directory, char_type const *searchSpec, recls_uint32_t flags);

/// @}

/// \name Iteration
/// @{
public:
    /// Begins the iteration
    ///
    /// \return An iterator representing the start of the sequence
    const_iterator  begin() const;
    /// Ends the iteration
    ///
    /// \return An iterator representing the end of the sequence
    const_iterator  end() const;

/// @}

/// \name State
/// @{
public:
    /// Returns the number of items in the sequence
    size_type           size() const;
    /// Indicates whether the sequence is empty
    recls_bool_t        empty() const;
    /// Returns the maximum number of items in the sequence
    static size_type    max_size();

/// @}

/// \name Members
/// @{
private:
    friend class basic_search_sequence_value_type<C, T>;
    friend class basic_search_sequence_const_iterator<C, T, value_type>;

    char_type       m_directory[_MAX_DIR + 1];
    char_type       m_search[_MAX_PATH + 1];
    recls_uint32_t  m_flags;

/// @}

// Not to be implemented
private:
    basic_search_sequence(class_type const &);
    basic_search_sequence const &operator =(class_type const &);
};

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs for commonly encountered types
 */

/// Instantiation of the basic_search_sequence template for the ANSI character type \c char
typedef basic_search_sequence<char, reclstl_traits<char> >          search_sequence_a;
/// Instantiation of the basic_search_sequence template for the Unicode character type \c wchar_t
typedef basic_search_sequence<wchar_t, reclstl_traits<wchar_t> >    search_sequence_w;

/* ////////////////////////////////////////////////////////////////////////// */

// class basic_search_sequence_value_type
/// Value type for the basic_search_sequence
///
/// \ingroup group_recls_stl
template<   typename C
        ,   typename T
        >
class basic_search_sequence_value_type
{
/// \name Types
/// @{
public:
    /// The character type
    typedef C                                                       char_type;
    /// The traits type
    typedef T                                                       traits_type;
    /// The current parameterisation of the type
    typedef basic_search_sequence_value_type<C, T>                  class_type;
    /// The find-data type
    typedef ss_typename_type_k traits_type::entry_type              entry_type;
    /// The directory parts sequence type
    typedef ss_typename_type_k traits_type::directory_parts_type    directory_parts_type;

/// @}

/// \name Construction
/// @{
public:
    /// Default constructor
    basic_search_sequence_value_type();
    basic_search_sequence_value_type(class_type const &rhs);
private:
    basic_search_sequence_value_type(entry_type info)
        : m_info(info)
    {}
public:
    /// Destructor
    ~basic_search_sequence_value_type();

    /// Copy assignment operator
    class_type &operator =(class_type const &rhs);

/// @}

/// \name Attributes
/// @{
public:
    /// Returns the full path of the item
    string_t                get_path() const;
#ifdef RECLS_PLATFORM_API_WIN32
    char_type               get_drive() const;
#endif /* RECLS_PLATFORM_API_WIN32 */
    /// Returns the directory of the item
    string_t                get_directory() const;
    /// Returns the directory (+ drive) of the item
    string_t                get_directory_path() const;
    /// Returns a sequence of directory parts of the item
    directory_parts_type    get_directory_parts() const;
    /// Returns the file (file name + extension) of the item
    string_t                get_file() const;
    /// Returns the short form of the filename of the item
    string_t                get_short_file() const;
    /// Returns the filename of the item
    string_t                get_filename() const;
    /// Returns the file extension of the item
    string_t                get_fileext() const;

    recls_time_t            get_creation_time() const;
    recls_time_t            get_modification_time() const;
    recls_time_t            get_last_access_time() const;
    recls_time_t            get_last_status_change_time() const;

    recls_filesize_t        get_size() const;

    recls_bool_t            is_readonly() const;
    recls_bool_t            is_directory() const;
    recls_bool_t            is_link() const;

    /// A C-string to the full path
    char_type const         *c_str() const;

    /// Returns a non-mutating reference to find-data
    entry_type const        &get_entry() const;

/// @}

// Members
private:
    friend class basic_search_sequence_const_iterator<C, T, class_type>;

    entry_type  m_info;
};

// class basic_search_sequence_const_iterator
/// Iterator type for the basic_search_sequence supporting the Input Iterator concept
///
/// \ingroup group_recls_stl
template<   typename C
        ,   typename T
        ,   typename V
        >
class basic_search_sequence_const_iterator
    : public stlsoft_ns_qual(iterator_base)<stlsoft_ns_qual_std(input_iterator_tag), V, ptrdiff_t, void, V>
{
public:
    /// The character type
    typedef C                                               char_type;
    /// The traits type
    typedef T                                               traits_type;
    /// The value type
    typedef V                                               value_type;
    /// The current parameterisation of the type
    typedef basic_search_sequence_const_iterator<C, T, V>   class_type;
    typedef stlsoft_define_move_rhs_type(class_type)        rhs_type;
    /// The find-data type
    typedef ss_typename_type_k traits_type::entry_type      entry_type;
private:
    typedef basic_search_sequence<C, T>                     sequence_type;

private:
    explicit basic_search_sequence_const_iterator(hrecls_t hSrch)
        : m_hSrch(hSrch)
    {}
public:
    /// Default constructor
    basic_search_sequence_const_iterator();
    /// <a href = "http://synesis.com.au/resources/articles/cpp/movectors.pdf">Move constructor</a>
    basic_search_sequence_const_iterator(rhs_type rhs);
    /// Destructor
    ~basic_search_sequence_const_iterator();

public:
    /// Pre-increment operator
    class_type &operator ++();
    /// Post-increment operator
    void operator ++(int);
    /// Dereference to return the value at the current position
    const value_type operator *() const;
    /// Evaluates whether \c this and \c rhs are equivalent
    recls_bool_t operator ==(class_type const &rhs) const;
    /// Evaluates whether \c this and \c rhs are not equivalent
    recls_bool_t operator !=(class_type const &rhs) const;

// Members
private:
    friend class basic_search_sequence<C, T>;

    hrecls_t    m_hSrch;

// Not to be implemented
private:
    basic_search_sequence_const_iterator &operator =(class_type const &rhs);
};

///////////////////////////////////////////////////////////////////////////////
// Shims

/// \brief is_empty shim
/// \ingroup group_recls_stl
///
/// This returns a non-zero value if the given sequence is empty.
/// \param s The sequence whose state is to be tested
///
/// \note This fits the <a href = "http://stlsoft.org/help/main.html">STLSoft</a> shims 
/// concept, described in <a href = "http://synesis.com.au/resources/articles/cpp/shims.pdf">this</a>
/// Synesis Software White Paper, and featured in the article "<a href = "http://www.cuj.com/documents/s=8681/cuj0308wilson/">Generalised String Manipulation: Access Shims and Type-tunnelling</a>",
/// in the August 2003 issue of <a href = "http://www.cuj.com">C/C++ User's Journal</a>.
template <typename C, typename T>
inline recls_bool_t is_empty(basic_search_sequence<C, T> const &s)
{
    return s.empty();
}

/// \brief c_str_ptr shim
/// \ingroup group_recls_stl
///
/// This returns a non-null null-terminated C-string of the given value's full path.
/// \param v The basic_search_sequence_value_type instance whose path is to be returned
///
/// \note This fits the <a href = "http://stlsoft.org/help/main.html">STLSoft</a> shims 
/// concept, described in <a href = "http://synesis.com.au/resources/articles/cpp/shims.pdf">this</a>
/// Synesis Software White Paper, and featured in the article "<a href = "http://www.cuj.com/documents/s=8681/cuj0308wilson/">Generalised String Manipulation: Access Shims and Type-tunnelling</a>",
/// in the August 2003 issue of <a href = "http://www.cuj.com">C/C++ User's Journal</a>.
template <typename C, typename T>
#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::char_type const *c_str_ptr(basic_search_sequence_value_type<C, T> const &v)
#else
inline C const *c_str_ptr(basic_search_sequence_value_type<C, T> const &v)
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */
{
    return v.c_str();
}

///////////////////////////////////////////////////////////////////////////////
// Implementation

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION

// basic_search_sequence

// Construction
template <typename C, typename T>
inline basic_search_sequence<C, T>::basic_search_sequence(char_type const *searchSpec, recls_uint32_t flags)
    : m_flags(flags)
{
    m_directory[0] = '\0';
    traits_type::str_copy(m_search, searchSpec);
}

template <typename C, typename T>
inline basic_search_sequence<C, T>::basic_search_sequence(char_type const *directory, char_type const * searchSpec, recls_uint32_t flags)
    : m_flags(flags)
{
    traits_type::str_copy(m_directory, directory);
    traits_type::str_copy(m_search, searchSpec);
}

// Iteration
template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence<C, T>::const_iterator basic_search_sequence<C, T>::begin() const
{
    return const_iterator(traits_type::Search(m_directory, m_search, m_flags));
}

template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence<C, T>::const_iterator basic_search_sequence<C, T>::end() const
{
    return const_iterator();
}

// State
template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence<C, T>::size_type basic_search_sequence<C, T>::size() const
{
    const_iterator  b   =   begin();
    const_iterator  e   =   end();
    size_type       c   =   0;

    for(; b != e; ++b)
    {
        ++c;
    }

    return c;
}

template <typename C, typename T>
inline recls_bool_t basic_search_sequence<C, T>::empty() const
{
    return begin() == end();
}

template <typename C, typename T>
inline /* static */ ss_typename_type_k basic_search_sequence<C, T>::size_type basic_search_sequence<C, T>::max_size()
{
    return static_cast<size_type>(-1);
}

// basic_search_sequence_value_type

template <typename C, typename T>
inline basic_search_sequence_value_type<C, T>::basic_search_sequence_value_type()
    : m_info(NULL)
{}

template <typename C, typename T>
inline basic_search_sequence_value_type<C, T>::basic_search_sequence_value_type(ss_typename_type_k basic_search_sequence_value_type<C, T>::class_type const &rhs)
    : m_info(traits_type::CopyDetails(rhs.m_info))
{}

template <typename C, typename T>
inline basic_search_sequence_value_type<C, T>::~basic_search_sequence_value_type()
{
    if(NULL != m_info)
    {
        traits_type::CloseDetails(m_info);
    }
}

template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::class_type &basic_search_sequence_value_type<C, T>::operator =(ss_typename_type_k basic_search_sequence_value_type<C, T>::class_type const &rhs)
{
    if(NULL != m_info)
    {
        traits_type::CloseDetails(m_info);
    }

    m_info = traits_type::CopyDetails(rhs.m_info);

    return *this;
}

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_path() const
{
    recls_assert(NULL != m_info);

    return string_t(m_info->path.begin, m_info->path.end);
}

#ifdef RECLS_PLATFORM_API_WIN32
template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::char_type basic_search_sequence_value_type<C, T>::get_drive() const
{
    char    chDrive;

    return (Recls_GetDriveProperty(m_info, &chDrive), chDrive);
}
#endif /* RECLS_PLATFORM_API_WIN32 */

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_directory() const
{
    recls_assert(NULL != m_info);

    return string_t(m_info->directory.begin, m_info->directory.end);
}

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_directory_path() const
{
    recls_assert(NULL != m_info);

    return string_t(m_info->path.begin, m_info->directory.end);
}

template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::directory_parts_type basic_search_sequence_value_type<C, T>::get_directory_parts() const
{
    recls_assert(NULL != m_info);

    return directory_parts_type(m_info->directoryParts.begin, m_info->directoryParts.end);
}

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_file() const
{
    recls_assert(NULL != m_info);

    return string_t(m_info->fileName.begin, m_info->fileExt.end);
}

#if 0
template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_short_file() const
{
    return m_info.cAlternateFileName[0] != '\0' ? m_info.cAlternateFileName : m_info.cFileName;
}
#endif /* 0 */

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_filename() const
{
    recls_assert(NULL != m_info);

    return string_t(m_info->fileName.begin, m_info->fileName.end);
}

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_fileext() const
{
    recls_assert(NULL != m_info);

    return string_t(m_info->fileExt.begin, m_info->fileExt.end);
}

template <typename C, typename T>
inline recls_time_t basic_search_sequence_value_type<C, T>::get_creation_time() const
{
    recls_assert(NULL != m_info);

    return Recls_GetCreationTime(m_info);
}

template <typename C, typename T>
inline recls_time_t basic_search_sequence_value_type<C, T>::get_modification_time() const
{
    recls_assert(NULL != m_info);

    return Recls_GetModificationTime(m_info);
}

template <typename C, typename T>
inline recls_time_t basic_search_sequence_value_type<C, T>::get_last_access_time() const
{
    recls_assert(NULL != m_info);

    return Recls_GetLastAccessTime(m_info);
}

template <typename C, typename T>
inline recls_time_t basic_search_sequence_value_type<C, T>::get_last_status_change_time() const
{
    recls_assert(NULL != m_info);

    return Recls_GetLastStatusChangeTime(m_info);
}

template <typename C, typename T>
inline recls_filesize_t basic_search_sequence_value_type<C, T>::get_size() const
{
    recls_assert(NULL != m_info);

    return m_info->size;
}

template <typename C, typename T>
inline recls_bool_t basic_search_sequence_value_type<C, T>::is_readonly() const
{
    recls_assert(NULL != m_info);

    return Recls_IsFileReadOnly(m_info);
}

template <typename C, typename T>
inline recls_bool_t basic_search_sequence_value_type<C, T>::is_directory() const
{
    recls_assert(NULL != m_info);

    return Recls_IsFileDirectory(m_info);
}

template <typename C, typename T>
inline recls_bool_t basic_search_sequence_value_type<C, T>::is_link() const
{
    recls_assert(NULL != m_info);

    return Recls_IsFileLink(m_info);
}

template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::char_type const * basic_search_sequence_value_type<C, T>::c_str() const
{
    recls_assert(NULL != m_info);

    return m_info->path.begin;
}

template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::entry_type const &basic_search_sequence_value_type<C, T>::get_entry() const
{
    return m_info;
}

// basic_search_sequence_const_iterator

template <typename C, typename T, typename V>
inline basic_search_sequence_const_iterator<C, T, V>::basic_search_sequence_const_iterator()
    : m_hSrch(NULL)
{}

template <typename C, typename T, typename V>
inline basic_search_sequence_const_iterator<C, T, V>::basic_search_sequence_const_iterator(rhs_type rhs)
    : m_hSrch(rhs.m_hSrch)
{
    move_lhs_from_rhs(rhs).m_hSrch = NULL;
}

template <typename C, typename T, typename V>
inline ss_typename_type_k basic_search_sequence_const_iterator<C, T, V>::class_type &basic_search_sequence_const_iterator<C, T, V>::operator =(typename basic_search_sequence_const_iterator<C, T, V>::class_type const &rhs)
{
    m_hSrch =   rhs.m_hSrch;

    return *this;
}

template <typename C, typename T, typename V>
inline basic_search_sequence_const_iterator<C, T, V>::~basic_search_sequence_const_iterator()
{
    if(NULL != m_hSrch)
    {
        Recls_SearchClose(m_hSrch);
    }
}

template <typename C, typename T, typename V>
inline ss_typename_type_k basic_search_sequence_const_iterator<C, T, V>::class_type &basic_search_sequence_const_iterator<C, T, V>::operator ++()
{
    recls_message_assert("Attempting to increment invalid iterator", NULL != m_hSrch);

    if(RECLS_FAILED(Recls_GetNext(m_hSrch)))
    {
        Recls_SearchClose(m_hSrch);
        m_hSrch = NULL;
    }

    return *this;
}

template <typename C, typename T, typename V>
inline void basic_search_sequence_const_iterator<C, T, V>::operator ++(int)
{
    operator ++();
}

template <typename C, typename T, typename V>
inline const ss_typename_type_k basic_search_sequence_const_iterator<C, T, V>::value_type basic_search_sequence_const_iterator<C, T, V>::operator *() const
{
    entry_type  info;

    if( m_hSrch != NULL &&
        RECLS_SUCCEEDED(traits_type::GetDetails(m_hSrch, &info)))
    {
        return value_type(info);
    }
    else
    {
        recls_message_assert("Dereferencing end()-valued iterator", 0);

        return value_type();
    }
}

template <typename C, typename T, typename V>
inline recls_bool_t basic_search_sequence_const_iterator<C, T, V>::operator ==(class_type const &rhs) const
{
    return (m_hSrch == NULL) && (rhs.m_hSrch == NULL);
}

template <typename C, typename T, typename V>
inline recls_bool_t basic_search_sequence_const_iterator<C, T, V>::operator !=(class_type const &rhs) const
{
    return ! operator ==(rhs);
}

#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////// */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace stl */
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* RECLS_INCL_H_WINSTL_SEARCH_SEQUENCE */

/* ////////////////////////////////////////////////////////////////////////// */
