/* /////////////////////////////////////////////////////////////////////////////
 * File:        SearchProcess_test.c
 *
 * Purpose:     Implementation file for the SearchProcess_test project.
 *
 * Created:     30th September 2003
 * Updated:     13th October 2003
 *
 * Status:      Wizard-generated
 *
 * License:     (Licensed under the Synesis Software Open License)
 *
 *              Copyright (C) 1999-2003, Synesis Software Pty Ltd.
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/software
 *
 *              email:      software@synesis.com.au
 *
 *              This source code is placed into the public domain 2003
 *              by Synesis Software Pty Ltd. There are no restrictions
 *              whatsoever to your use of the software. 
 *
 *              This source code is provided by Synesis Software Pty Ltd "as is"
 *              and any warranties, whether expressed or implied, including, but
 *              not limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software Pty Ltd be liable for any direct,
 *              indirect, incidental, special, exemplary, or consequential
 *              damages (including, but not limited to, procurement of
 *              substitute goods or services; loss of use, data, or profits; or
 *              business interruption) however caused and on any theory of
 *              liability, whether in contract, strict liability, or tort
 *              (including negligence or otherwise) arising in any way out of
 *              the use of this software, even if advised of the possibility of
 *              such damage. 
 *
 *              Neither the name of Synesis Software Pty Ltd nor the names of
 *              any subdivisions, employees or agents of Synesis Software Pty
 *              Ltd, nor the names of any other contributors to this software
 *              may be used to endorse or promote products derived from this
 *              software without specific prior written permission. 
 *
 * ////////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <string.h>

#include <recls.h>
#include <recls_assert.h>

#if RECLS_VER < RECLS_VER_1_1_1
# error This file now requires version 1.1.1 or later of the recls C API
#endif /* RECLS_VER < 1.1.1 */

/* /////////////////////////////////////////////////////////////////////////////
 * Macros
 */

#ifndef NUM_ELEMENTS
# define NUM_ELEMENTS(x)        (sizeof(x) / sizeof((x)[0]))
#endif /* !NUM_ELEMENTS */

/* /////////////////////////////////////////////////////////////////////////////
 * Forward declarations
 */

void usage(int bExit);
int RECLS_CALLCONV_DEFAULT EntryFunc(recls_info_t info, recls_process_fn_param_t param);

/* /////////////////////////////////////////////////////////////////////////////
 * Main
 */

/** Program main */
int main(int argc, char *argv[])
{
    int                 iRet    =   0;
    int                 i;
    char const          *pattern    =   NULL;
    char const          *rootDir    =   NULL;
    recls_rc_t          rc;
    recls_uint32_t      flags       =   RECLS_F_RECURSIVE;

	for(i = 1; i < argc; ++i)
	{
		const char	*arg	=	argv[i];

		if(arg[0] == '-')
		{
			if(arg[1] == '-')
			{
				/* -- arguments */
			}
			else
			{
                /* - arguments */
                switch(arg[1])
                {
                    case    'R':    /* Do not recurse */
                        flags &= ~(RECLS_F_RECURSIVE);
                        break;
                    case    'p':    /* Show directory parts */
                        flags |= RECLS_F_DIRECTORY_PARTS;
                        break;
                    case    'f':    /* Find files */
                        flags |= RECLS_F_FILES;
                        break;
                    case    'd':    /* Find directories */
                        flags |= RECLS_F_DIRECTORIES;
                        break;
                    default:
                        usage(1);
                        break;
				}
			}
		}
		else
		{
            /* other arguments */
            if(NULL == pattern)
            {
                pattern = arg;
            }
            else if(NULL == rootDir)
            {
                rootDir = arg;
            }
            else
            {
                usage(1);
            }
		}
	}

    /* Search for files if neither files or directories specified/
     *
     * Even though this is not necessary, because the recls API provides the
     * same interpretation, it's best to be explicit.
     */
    if(0 == (flags & (RECLS_F_FILES | RECLS_F_DIRECTORIES)))
    {
        flags |= RECLS_F_FILES;
    }

    if(NULL == pattern)
    {
#if defined(RECLS_PLATFORM_IS_WIN32)
        pattern = "*.*";
#else
        pattern = "*";
#endif /* RECLS_PLATFORM_IS_WIN32 */
    }

    if(NULL == rootDir)
    {
        rootDir = ".";
    }

	/* Now just run the search with the given conditions. */
	rc = Recls_SearchProcess(rootDir, pattern, flags, EntryFunc, (recls_process_fn_param_t)stdout);

    if(RECLS_FAILED(rc))
    {
        char    err[100];

        Recls_GetErrorString(rc, err, 100);

        fprintf(stderr, "Failed to start search, with pattern \"%s\"; recls error: %s\n", pattern, err);
    }

#ifdef _MSC_VER
    /* Check that there are no leaks */
    _CrtDumpMemoryLeaks();
#endif /* _MSC_VER */

    return iRet;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Function implementations
 */

/** Displays program usage */
void usage(int bExit)
{
    fprintf(stderr, "recls SearchProcess Test Program: SearchProcess_test.exe\n\n");
    fprintf(stderr, "Usage: SearchProcess_test [-R] [<pattern>] [<root-dir>]\n");
    fprintf(stderr, "\t-R               -   does not recurse. (recursive search is the default)\n");
    fprintf(stderr, "\t<pattern>        -   search pattern, e.g. \"*.cpp\" (default is to search for all files)\n");
    fprintf(stderr, "\t<root-dir>       -   root directory of search; default is current working directory\n");

    if(bExit)
    {
        exit(1);
    }
}

/** The callback entry processor function. */
int RECLS_CALLCONV_DEFAULT EntryFunc(recls_info_t info, recls_process_fn_param_t param)
{
	FILE	*stream = (FILE*)param;
    int     i;
    int     off;
    int     extLen;
    int     cDirParts;
    char    path[RECLS_PATH_MAX];
#ifdef RECLS_PLATFORM_API_WIN32
    char    drive;
#endif /* RECLS_PLATFORM_API_WIN32 */
    char    dir[RECLS_PATH_MAX];
    char    dirPath[RECLS_PATH_MAX];
    char    file[RECLS_PATH_MAX];
    char    fileName[RECLS_PATH_MAX];
    char    fileExt[RECLS_PATH_MAX];

    char    pathCheck[RECLS_PATH_MAX];
    char    dirCheck[RECLS_PATH_MAX];
    char    fileCheck[RECLS_PATH_MAX];

    pathCheck[0] = '\0';
    dirCheck[0] = '\0';
    fileCheck[0] = '\0';

    Recls_GetPathProperty(info, path, NUM_ELEMENTS(path));
    fprintf(stream, "  %s\n", path);
	Recls_GetDirectoryPathProperty(info, dirPath, NUM_ELEMENTS(dirPath));
    fprintf(stream, "  %s\n", dirPath);
#ifdef RECLS_PLATFORM_API_WIN32
    Recls_GetDriveProperty(info, &drive);
    if(isupper(path[0]))
    {
        drive = (char)toupper(drive);
    }
    else
    {
        drive = (char)tolower(drive);
    }
    fprintf(stream, "  %c\n", drive);
    sprintf(pathCheck, "%c:", drive);
#endif /* RECLS_PLATFORM_API_WIN32 */
    Recls_GetDirectoryProperty(info, dir, NUM_ELEMENTS(dir));
    fprintf(stream, "    %s\n", dir);

	off = 0;
    for(i = 0, cDirParts = Recls_GetDirectoryPartProperty(info, -1, NULL, 0); i < cDirParts; ++i)
    {
        char    dirPart[RECLS_PATH_MAX];

        off += Recls_GetDirectoryPartProperty(info, i, dirPart, NUM_ELEMENTS(dirPart));
        fprintf(stream, "    %*s\n", off, dirPart);

        strcat(pathCheck, dirPart);
        strcat(dirCheck, dirPart);
    }
	if(0 == cDirParts)
    {
        off += strlen(dir);
		strcat(pathCheck, dir);
    }

    off +=   Recls_GetFileProperty(info, file, NUM_ELEMENTS(file));
             Recls_GetFileNameProperty(info, fileName, NUM_ELEMENTS(fileName));
    extLen = Recls_GetFileExtProperty(info, fileExt, NUM_ELEMENTS(fileExt));
    fprintf(stream, "    %*s\n", off, file);
    fprintf(stream, "    %*s\n", off - (1 + extLen), fileName);
    fprintf(stream, "    %*s\n", off, fileExt);
    strcpy(fileCheck, fileName);
    if(0 < extLen)
    {
        strcat(fileCheck, ".");
        strcat(fileCheck, fileExt);
    }

    strcat(pathCheck, file);

    /* Now validate the components */
#if defined(RECLS_PLATFORM_IS_WIN32) && \
    !defined(RECLS_PLATFORM_API_WIN32)
            recls_assert(0 == strcmp(path + 2, pathCheck));
            if(0 != strcmp(path + 2, pathCheck))
#else /* ? RECLS_PLATFORM_API_WIN32 */
            recls_assert(0 == strcmp(path, pathCheck));
            if(0 != strcmp(path, pathCheck))
#endif /* RECLS_PLATFORM_API_WIN32 || !RECLS_PLATFORM_IS_WIN32 */
    {
        fprintf(stderr, "Path is different from path components\n\tpath:  %s\n\tparts: %s\n\n", path, pathCheck);

        abort();
    }
#if defined(RECLS_PLATFORM_API_WIN32)
            recls_assert(dirPath[0] == drive && 0 == strcmp(dir, dirPath + 2));
            if( dirPath[0] != drive ||
                0 != strcmp(dir, dirPath + 2))
#elif defined(RECLS_PLATFORM_IS_WIN32)
            recls_assert(0 == strcmp(dir, dirPath + 2));
            if(0 != strcmp(dir, dirPath + 2))
#else
            recls_assert(dirPath == dir);
            if(dirPath != dir)
#endif /* RECLS_PLATFORM_API_WIN32 || !RECLS_PLATFORM_IS_WIN32 */
    {
        fprintf(stderr, "DirectoryPath is different from Directory\n\tdirPath:  %s\n\tdir: %s\n\n", dirPath, dir);

        abort();
    }
    if(0 != cDirParts)
    {
        recls_assert(0 == strcmp(dir, dirCheck));
        if(0 != strcmp(dir, dirCheck))
        {
            fprintf(stderr, "Directory is different from directory components\n\tpath:  %s\n\tparts: %s\n\n", dir, dirCheck);

            abort();
        }
    }
    recls_assert(0 == strcmp(file, fileCheck));
    if(0 != strcmp(file, fileCheck))
    {
        fprintf(stderr, "File is different from file components\n\tpath:  %s\n\tparts: %s\n\n", file, fileCheck);

        abort();
    }

    if(Recls_IsFileReadOnly(info))
    {
        fprintf(stream, "    - Read-only\n");
    }
    if(Recls_IsFileDirectory(info))
    {
        fprintf(stream, "    - Directory\n");
    }
    else
    {
        fprintf(stream, "    - File\n");
    }
    if(Recls_IsFileLink(info))
    {
        fprintf(stream, "    - Link\n");
    }

	return 1;
}

/* //////////////////////////////////////////////////////////////////////////
 * End of file
 */
